#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "hash.h"
#include "sha1.h"

#if !defined(O_SYNC) && defined(O_FSYNC)
# define	O_SYNC	O_FSYNC
#endif
#if defined(O_SYNC) && !defined(O_FSYNC)
# define	O_FSYNC	O_SYNC
#endif


htable_t	htable[] = {
	{ "SHA",	".sha1",	HASH_SHA_1 },
	{ "SHA1",	".sha1",	HASH_SHA_1 },
	{ "SHA-1",	".sha1",	HASH_SHA_1 },
	{ NULL,		NULL,		0 }
};

static hashtype_t hashes[] = {
	{ HASH_SHA_1, "SHA-1", NULL, NULL, NULL, SHA1_DIGEST_SZ, SHA1Create,
	  SHA1Destroy, (void *)SHA1Init, (void *)SHA1Update,
	  (void *)SHA1Final },
	{ HASH_NONE, "NONE", NULL, NULL, NULL, 0, NULL, NULL, NULL, NULL }
};


htable_t *
hash_lookup(type)
	char *type;
{
	int i;

	for (i = 0; htable[i].ht_name; i++)
		if (strcasecmp(htable[i].ht_name, type) == 0)
			return htable + i;
	return NULL;
}


htable_t *
hash_lookuptype(type)
	int type;
{
	int i;

	for (i = 0; htable[i].ht_name; i++)
		if (htable[i].ht_type == type)
			return htable + i;
	return NULL;
}


hashtype_t *
hash_find(type)
	int type;
{
	int i;

	for (i = 0; hashes[i].ht_type != HASH_NONE; i++)
		if (hashes[i].ht_type == type) {
			if (hashes[i].ht_tab == NULL)
				hashes[i].ht_tab = hash_lookuptype(type);
			return hashes + i;
		}
	return NULL;
}


char *
hash_getsalt(buf, bufsz)
	char *buf;
	size_t bufsz;
{
	struct timeval tv;
	int fd, hv, n;
	char *s;

	fd = open(HASH_RANDOM, O_RDONLY);
	if (fd >= 0) {
		/*
		 * Urgh, some /dev/random's don't supply a very continuous
		 * stream of bits...
		 */
		for (s = buf; s - buf < bufsz; ) {
			n = read(fd, s, 1);
			if (n == -1)
				break;
			s += n;
		}

		close(fd);
		if (s - buf == bufsz)
			return buf;
	}

	/*
	 * XXX - This is not exactly strong, but it'll have to do for now
	 * as I don't want to write a real pseudo-random generator.
	 */
	(void) gettimeofday(&tv, NULL);
	hv = tv.tv_sec * tv.tv_usec;
	(void) gettimeofday(&tv, NULL);
	hv *= tv.tv_usec ^ hv;
	srandom(hv);
	for (s = buf; s - buf < bufsz; s++) {
		hv = random();
		*s = hv & 0x0f;
		hv = random();
		*s |= hv & 0xf0;
	}
	return buf;
}


void
hash_savesalt(file, salt, slen)
	char *file;
	char *salt;
	size_t slen;
{
	struct stat sb;
	int fd;

	if (file == NULL)
		file = NSYSLOG_SALT;
	fd = open(NSYSLOG_SALT, O_WRONLY|O_SYNC|O_CREAT, 0600);
	if (fd >= 0) {
		if (write(fd, salt, slen) != slen)
			perror("hash_writesalt");
		close(fd);
	}
}


/*
 * No use for this function ?
 */
void
hash_readsalt(file, salt, slen)
	char *file;
	char *salt;
	size_t slen;
{
	int fd;

	if (file == NULL)
		file = NSYSLOG_SALT;
	fd = open(NSYSLOG_SALT, O_RDONLY);
	if (fd >= 0) {
		if (read(fd, salt, slen) != slen)
			perror("hash_readsalt");
		close(fd);
	}
}


static void
hash_updatedbuf(hptr, dbuf)
	hashtype_t *hptr;
	char *dbuf;
{
	static time_t last_H0 = 0;
	time_t now;

	if (hptr == NULL)
		return;

	hash_init(hptr);
	hash_update(hptr, dbuf, hptr->ht_dsize);
	hash_final(hptr, dbuf);

	if (last_H0 == 0) {
		hash_flush();
		last_H0 = time(NULL);
	} else {
		now = time(NULL);
		if (now - last_H0 > HASH_WRITE_INTERVAL) {
			last_H0 = now;
			hash_flush();
		}
	}
}


/*
 * given the path to the normal log file, determine what the path is for the
 * file with H0 in it and attempt to read in its contents.
 */
int
hash_openH0(hashdir, path, mode)
	char *hashdir, *path;
	int mode;
{
	char *hpath, *hdir;
	int hfd = -1, len;

	if (*path == '/')
		len = strlen(path);
	else {
		hdir = hashdir;
		if (hdir == NULL)
			hdir = HASH_LOGDIR;
 
		len = strlen(hdir) + strlen(path) + 2;
	}
	len += strlen(HASH_H0_EXT) + 2;

	hpath = (char *)malloc(len);
	*hpath = '\0';

	if (*path == '/')
		*hpath = '\0';
	else if (*hdir != '\0') {
		strcpy(hpath, hdir);
		strcat(hpath, "/");
	}
	strcat(hpath, path);
	if (strcmp(path + strlen(path) - strlen(HASH_H0_EXT), HASH_H0_EXT))
		strcat(hpath, HASH_H0_EXT);
 
	if (mode == O_WRONLY)
		hfd = open(hpath, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC, 0600);
	else if (mode == O_RDONLY)
		hfd = open(hpath, O_RDONLY, 0600);
	else
		hfd = -1;
	if (hfd == -1)
		syserr("couldn't open %s for %sing: %m", hpath,
			mode == O_RDONLY ? "read" : "writ");
	else {
		(void) fchmod(hfd, 0400);
		(void) fcntl(hfd, F_SETFD, fcntl(hfd, F_GETFD, 0) | FD_CLOEXEC);
	}
	free(hpath);
	return hfd;

}


/*
 * save the current hash `password' to a file.
 */
void
hash_writeH0(hptr, path, dbuf, lcnt, lcntsz)
	hashtype_t *hptr;
	char *path, *dbuf;
	u_long *lcnt;
	size_t lcntsz;
{
	int fd;

	fd = hash_openH0(NULL, path, O_WRONLY);
	if (fd == -1)
		return;

	if (write(fd, dbuf, hptr->ht_dsize) != hptr->ht_dsize)
		syserr("write to hash file failed: %m");
	else if (write(fd, (char *)lcnt, lcntsz) != lcntsz)
		syserr("write to hash file failed: %m");
	close(fd);
}


hashtype_t *
hash_get(type)
	int type;
{
	hashtype_t *ht, *htf;

	htf = hash_find(type);
	if (htf != NULL) {
		if (!(NEW(ht, NEW_HASH)))
			return NULL;
		bcopy((char *)htf, (char *)ht, sizeof(*ht));
		(*ht->ht_create)(ht);
		return ht;
	}
	return NULL;
}


hashtype_t *
hash_getnext(type)
	int type;
{
	hashtype_t *ht, *htf;

	htf = hash_find(type);
	if (htf != NULL) {
		htf++;
		htf = hash_find(htf->ht_type);
	}
	if (htf != NULL) {
		if (!(NEW(ht, NEW_HASH)))
			return NULL;
		bcopy((char *)htf, (char *)ht, sizeof(*ht));
		(*ht->ht_create)(ht);
		return ht;
	}
	return NULL;
}


void
hash_free(ht)
	hashtype_t *ht;
{
	(*ht->ht_destroy)(ht);
	free(ht);
}


void
hash_init(hptr)
	hashtype_t *hptr;
{
	(*hptr->ht_init)(hptr->ht_buf);
	hptr->ht_digest = malloc(hptr->ht_dsize);
}


void
hash_update(hptr, buf, len)
	hashtype_t *hptr;
	char *buf;
	size_t len;
{
	(*hptr->ht_update)(hptr->ht_buf, buf, len);
}


char *
hash_final(hptr, digest)
	hashtype_t *hptr;
	char *digest;
{
	if (digest == NULL)
		digest = (char *)malloc(hptr->ht_dsize);
	(*hptr->ht_final)(digest, hptr->ht_buf);
	return digest;
}




int
hash_openfile(hdir, hpath, ldir, path, hptr)
	char *hdir, *hpath, *ldir, *path;
	hashtype_t *hptr;
{
	int hfd = -1, len;
	char *ext;

	ext = hptr->ht_tab->ht_ext;
	if (hpath == NULL) {
		/*
		 * If the path starts at /, then only add the extension,
		 * otherwise, generate the complete path for opening.
		 */
		if (*path == '/')
			len = strlen(path) + strlen(ext) + 2;
		else {
			if (!hdir && ldir)
				hdir = ldir;
			else if (!hdir)
				hdir = "";
			len = strlen(hdir) + strlen(path) + strlen(ext) +4;
		}
		hpath = (char *)malloc(len);

		if (*path == '/')
			*hpath = '\0';
		else if ((hdir != NULL) && (*hdir != '\0')) {
			strcpy(hpath, hdir);
			strcat(hpath, "/");
		}
		strcat(hpath, path);
		strcat(hpath, ext);
	}
	hfd = open(hpath, O_RDWR|O_APPEND|O_CREAT, 0600);
	if (hfd != -1)
		(void) fcntl(hfd, F_SETFD, fcntl(hfd, F_GETFD, 0) | FD_CLOEXEC);
	free(hpath);
	return hfd;
}
